#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from src.client.client import LiquidSoapClient


class LiquidSoapPlayerClient(LiquidSoapClient):

    # TODO Refactor class: https://gitlab.servus.at/aura/engine/-/issues/65

    #
    # Mixer
    #

    # def mixer(self, mixer_id, command, *args):
    #     if command == "status":
    #         return self.mixer_status(mixer_id, *args)

    #     if command == "inputs":
    #         return self.mixer_inputs(mixer_id)

    #     if command == "volume":
    #         return self.mixer_volume(mixer_id, *args)

    #     if command == "select":
    #         if len(args) == 2:
    #             return self.mixer_select(mixer_id, args[0], args[1])

    #     if command == "activate":
    #         if len(args) == 2:
    #             return self.mixer_activate(mixer_id, args[0], args[1])

    #     return "LiquidSoapPlayerClient does not understand mixer."+command+str(args)

    def mixer_inputs(self, mixer_id):
        # send command
        self.command(mixer_id, "inputs")

        # convert to list and return it
        return self.message.strip().split(" ")

    def mixer_status(self, mixer_id, pos=""):
        """
        Get state of a source in the mixer
        @type    pos:       string
        @param   pos:       Mixerposition
        @rtype:  string
        @return: Response from LiquidSoap
        """
        self.command(mixer_id, "status", str(pos))
        return self.message

    # def input_status(self, input_id):
    #     """
    #     Retrieves the status of the given input

    #     Args:
    #         input_id (_type_): _description_
    #     """
    #     self.command(input_id, "status")
    #     return self.message

    def mixer_volume(self, mixer_id, pos, volume):
        """
        Sets some mixer channel to the given volume

        Args:
            pos (Integer):  The channel number
            volume (Integer): The volume

        Returns:
            (String):   Liquidsoap server response
        """
        self.command(mixer_id, "volume", str(pos) + " " + str(volume))
        return self.message

    def mixer_select(self, mixer_id, pos, select):
        """
        Selects some mixer channel or vice versa.

        Args:
            pos (Integer):  The channel number
            select (Boolean): Select or deselect

        Returns:
            (String):   Liquidsoap server response
        """
        self.command(mixer_id, "select", str(pos) + " " + str(select).lower())
        return self.message

    def mixer_activate(self, mixer_id, pos, activate):
        """
        Selects some mixer channel and increases the volume to 100 or vice versa.

        Args:
            pos (Integer):  The channel number
            activate (Boolean): Activate or deactivate

        Returns:
            (String):   Liquidsoap server response
        """
        self.command(mixer_id, "activate", str(pos) + " " + str(activate).lower())
        return self.message

    #
    # Channel (general)
    #

    def set_track_metadata(self, channel, json_meta):
        """
        Sets additional metadata for the current track
        """
        self.command(channel, "set_track_metadata", json_meta)
        return self.message

    #
    # Queues
    #

    def queue_push(self, channel, uri):
        """
        Pushes the passed file URI to the `equeue` playlist channel.

        Args:
            channel (String):   Liquidsoap Source ID
            uri (String):       Path to the file
        """
        self.command(channel, "push", uri)
        return self.message

    def queue_seek(self, channel, duration):
        """
        Forward the playing `equeue` track/playlist of the given channel.

        Args:
            channel (String):   Liquidsoap Source ID
            duration (Integer): Seek duration ins seconds

        Returns:
            Liquidsoap server response
        """
        self.command(channel, "seek", str(duration))
        return self.message

    def queue_clear(self, channel):
        """
        Clears all `equeue` playlist entries of the given channel.

        Args:
            channel (String):   Liquidsoap Source ID
            duration (Integer): Seek duration ins seconds

        Returns:
            Liquidsoap server response
        """
        self.command(channel, "clear")
        return self.message

    #
    # Playlist
    #

    def playlist_uri_set(self, channel, uri):
        """
        Sets the URI of a playlist source.

        Args:
            channel (String):   Liquidsoap Source ID
            uri (String):       URI to the playlist file

        Returns:
            Liquidsoap server response
        """
        self.command(channel, "uri", uri)
        return self.message

    def playlist_uri_clear(self, channel):
        """
        Clears the URI of a playlist source.

        Args:
            channel (String):   Liquidsoap Source ID
            uri (String):       URI to the playlist file

        Returns:
            Liquidsoap server response
        """
        self.command(channel, "clear")
        return self.message

    #
    # Stream
    #

    def stream_set_url(self, channel, url):
        """
        Sets the URL on the given HTTP channel.
        """
        self.command(channel, "url", url)
        return self.message

    def stream_start(self, channel):
        """
        Starts the HTTP stream set with `stream_set_url` on the given channel.
        """
        self.command(channel, "start")
        return self.message

    def stream_stop(self, channel):
        """
        Stops the HTTP stream on the given channel.
        """
        self.command(channel, "stop")
        return self.message

    def stream_status(self, channel):
        """
        Returns the status of the HTTP stream on the given channel.
        """
        self.command(channel, "status")
        return self.message

    #
    # General Entries
    #

    def entry_status(self, rid):
        """
        Retrieves the status of a given entry.

        Args:
            rid (String):   Resource ID (RID)

        Returns:
            Liquidsoap server response
        """
        self.command("request", "status", str(rid))
        return self.message

    #
    # Other
    #

    def uptime(self, command=""):  # no command will come
        """
        Retrieves how long the engine is running already.
        """
        return self.command("", "uptime")

    def version(self, command=""):  # no command will come
        """
        Retrieves the Liquidsoap version.
        """
        return self.command("", "version")

    def engine(self, command, *args):
        """
        Retrieves the state of all input and outputs.
        """
        if command == "state":
            return self.engine_state()
        if command == "update_config":
            return self.engine_update_config(str(args[0]))
        if command == "set_track_metadata":
            return self.engine_set_track_metadata(str(args[0]))

        return "LiquidSoapPlayerClient does not understand engine." + command + str(args)

    def engine_state(self):
        """
        Retrieves the state of all input and outputs.
        """
        self.command("aura_engine", "status")
        return self.message

    def engine_update_config(self, json_config):
        """
        Updates the config
        """
        self.command("aura_engine", "update_config", json_config)
        return self.message

    # def skip(self, namespace="playlist", pos=""):
    #     """
    #     Source skippen
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @type    pos:       string
    #     @param   pos:       Die Position - optional - Position des Channels vom Mixer benötigt
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('skip', namespace, pos)
    #     return self.message

    # def remove(self, pos, namespace="playlist"):
    #     """
    #     Track  aus der secondary_queue oder der Playlist entfernen
    #     @type    pos:       string
    #     @param   pos:       Die Position
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('remove', namespace, str(pos))
    #     return self.message

    # def insert(self, uri, pos='0', namespace="playlist"):
    #     """
    #     Track  einfügen
    #     @type    uri:       string
    #     @param   uri:       Uri einer Audiodatei
    #     @type    pos:       string
    #     @param   pos:       Die Position
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('insert', namespace, str(pos) + ' ' + uri)
    #     return self.message

    # def move(self, fromPos, toPos, namespace="playlist"):
    #     """
    #     Track  von Position fromPos nach Position toPos verschieben
    #     @type    fromPos:   string/int
    #     @param   fromPos:   Position des zu verschiebenden Tracks
    #     @type    toPos:     string
    #     @param   toPos:     Die Position zu der verschoben werden soll
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('move', namespace, str(fromPos) + ' ' + str(toPos))
    #     return self.message

    # def play(self, namespace="playlist"):
    #     """
    #     Source abspielen - funktioniert nur bei Playlist
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('play', namespace)
    #     return self.message

    # def pause(self, namespace="playlist"):
    #     """
    #     Source pausieren/stoppen - funktioniert nur bei Playlist
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('pause', namespace)
    #     return self.message

    # def flush(self, namespace="playlist"):
    #     """
    #     Playlist leeren
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('flush', namespace)
    #     return self.message

    # def playlistData(self):
    #     """
    #     Metadaten der Playlist ausgeben
    #     @rtype:  string
    #     @return: Ein Json-String
    #     """
    #     self.command('data', 'playlist')
    #     return self.message

    # def get_queue(self, namespace="ch1", queue='queue'):
    #     """
    #     Queue eines Kanals ausgeben
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @type    queue:     string
    #     @param   queue:    Name des queues (queue, primary_queue, secondary_queue)
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command(queue, namespace)
    #     return self.message

    # def loadPlaylist(self, uri, params="", namespace="playlist"):
    #     """
    #     Playlist laden
    #     @type    uri:       string
    #     @param   uri:       Uri einer Playlist im XSPF-Format
    #     @type    params:    string
    #     @param   params:    obsolete
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source - hier nur playlist
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('load', namespace, uri + params)
    #     return self.message

    # def currentTrack(self, namespace="request"):
    #     """
    #     Das oder die ID(s) der gerade abgespielten requests erhalten
    #     @type    namespace: string
    #     @param   namespace: Namespace der Source
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers (als String)
    #     """
    #     self.command('on_air', namespace)
    #     return self.message

    def volume(self, pos, volume, namespace="mixer"):
        """
        Lautstärke eines Kanals setzen
        @type    pos:       int/string
        @param   pos:       Die Position/ Nummer des Kanals (playlist=0)
        @type    volume:    int/string
        @param   volume:    Zahl von 1 -100
        @type    namespace: string
        @param   namespace: Namespace der Source (immer mixer)
        @rtype:  string
        @return: Die Antwort des Liquidsoap-Servers
        """
        self.command("volume", namespace, str(pos) + " " + str(volume))
        return self.message

    # def playlist_remaining(self):
    #     """
    #     Wie lange läuft der aktuelle Track der Playlist noch
    #     @rtype:  string
    #     @return: Die Antwort des Liquidsoap-Servers
    #     """
    #     self.command('remaining', 'playlist')
    #     return self.message

    # def list_channels(self):
    #     """
    #     Channels auflisten (Simple JSON)
    #     """
    #     # Liquidsoap Kommando

    #     channels = self.sendLqcCommand(self.lqc, 'mixer', 'inputs')

    #     if not isinstance(channels, list):
    #         self.error('02')
    #     elif len(channels) < 1:
    #         self.warning('01')
    #     else:
    #         self.success('00', channels)

    #     self.notifyClient()
